@@ -1,10 +1,12 @@
 // This file may be redistributed and modified only under the terms of
 // the GNU Lesser General Public License (See COPYING for details).
-// Copyright (C) 2000-2001 Stefanus Du Toit, Karsten-O. Laux, Alistair Riddoch
+// Copyright (C) 2000-2001 Stefanus Du Toit, Karsten-O. Laux and Al Riddoch
 
 #ifndef ATLAS_MESSAGE_ELEMENT_H
 #define ATLAS_MESSAGE_ELEMENT_H
 
+#include <Atlas/Exception.h>
+
 #include <string>
 #include <map>
 #include <vector>
@@ -12,7 +14,20 @@
 namespace Atlas { namespace Message {
 
 /// An exception class issued when the wrong type is requested in as().
-class WrongTypeException { };
+class WrongTypeException : public Atlas::Exception
+{
+  public:
+    WrongTypeException() : Atlas::Exception("Wrong Message::Element type") { }
+};
+
+class Element;
+
+typedef long IntType;
+typedef double FloatType;
+typedef void * PtrType;
+typedef std::string StringType;
+typedef std::map<std::string, Element> MapType;
+typedef std::vector<Element> ListType;
 
 /** Multi-type container
  *
@@ -22,10 +37,15 @@ class WrongTypeException { };
  *
  * Changes:
  *
+ *   2003/04/02   Al Riddcoh <alriddoch@zepler.org>
+ *                Add in some assignment operators for efficiency
+ *   2002/11/07   Al Riddcoh <alriddoch@zepler.org>
+ *                Changed the name to Element as Object is a stupid name
+ *                for a class.
  *   2000/08/05   Karsten-O. Laux <klaux@rhrk.uni-kl.de>
  *                Changed the members to pointers which only get created when
  *                really needed. This is a major speedup for passing
- *                Element as parameter or when copying it.
+ *                Object as parameter or when copying it.
  *                Because copying of unused members is omitted.  
  *                All pointers are stored as a union, so we save memory !
  *                Changed IntType to long and added convinience Constructors
@@ -37,82 +57,44 @@ class WrongTypeException { };
 class Element
 {
 public:
-    typedef long IntType;
-    typedef double FloatType;
-    typedef std::string StringType;
-    typedef std::map<std::string, Element> MapType;
-    typedef std::vector<Element> ListType;
-
     enum Type {
         TYPE_NONE,
         TYPE_INT,
         TYPE_FLOAT,
+        TYPE_PTR,
         TYPE_STRING,
         TYPE_MAP,
         TYPE_LIST
     };
 
+private:
+    // These are now legacy typedefs. New code should use the
+    // Atlas::Message::*Type versions.
+    typedef Atlas::Message::IntType IntType;
+    typedef Atlas::Message::FloatType FloatType;
+    typedef Atlas::Message::PtrType PtrType;
+    typedef Atlas::Message::StringType StringType;
+    typedef Atlas::Message::MapType MapType;
+    typedef Atlas::Message::ListType ListType;
+
+    /// Clear all values.
+    void clear(Type new_type = TYPE_NONE);
+
+public:
     /// Construct an empty object.
     Element()
       : t(TYPE_NONE)
     {
     }
 
-    /// Clear all values.
-    void clear()
-    {
-     switch(t) 
-        {
-      case TYPE_NONE:
-      case TYPE_INT:
-      case TYPE_FLOAT:
-        break;
-      case TYPE_STRING:
-        delete s;
-        break;
-      case TYPE_MAP:
-        delete m;
-        break;
-      case TYPE_LIST:
-        delete l;
-        break;
-      }
-     
-     t = TYPE_NONE;
-    }
-
     ///
-    virtual ~Element()
-      {
+    ~Element()
+    {
         clear();
-      }
+    }
 
     /// Copy an existing object.
-    Element(const Element& obj)
-      : t(obj.t)
-    {
-      switch(t) 
-        {
-      case TYPE_NONE:
-        break;
-      case TYPE_INT:
-        i = obj.i;
-        break;
-      case TYPE_FLOAT:
-        f = obj.f;
-        break;
-      case TYPE_STRING:
-        s = new StringType(*obj.s);
-        break;
-      case TYPE_MAP:
-        m = new MapType(*obj.m);
-        break;
-      case TYPE_LIST:
-        l = new ListType(*obj.l);
-        break;
-      }
-        
-    }
+    Element(const Element& obj);
 
     /// Set type to int, and value to v.
     Element(int v)
@@ -144,91 +126,167 @@ public:
     {
     }
 
+    /// Set type to PtrType, and value to v.
+    Element(PtrType v)
+      : t(TYPE_PTR), p(v)
+    {
+    }
+
     /// Set type to std::string, and value to v.
     Element(const char* v)
       : t(TYPE_STRING)
     {
       if(v)
-        s = new StringType(v);
+        s = new DataType<StringType>(v);
       else
-        s = new StringType();
+        s = new DataType<StringType>();
     }
 
     /// Set type to std::string, and value to v.
     Element(const StringType& v)
       : t(TYPE_STRING)
     {
-      s = new StringType(v);
+      s = new DataType<StringType>(v);
     }
     /// Set type to MapType, and value to v.
     Element(const MapType& v)
       : t(TYPE_MAP)
     {
-      m = new MapType(v);
+      m = new DataType<MapType>(v);
     }
     /// Set type to ListType, and value to v.
     Element(const ListType& v)
       : t(TYPE_LIST)
     {
-      l = new ListType(v);
+      l = new DataType<ListType>(v);
     }
 
     /// overload assignment operator !
-    Element& operator=(const Element& obj) 
+    Element& operator=(const Element& obj);
+
+    Element& operator=(int v) 
     {
-      //check for self assignment
-      if(&obj == this)
-        return *this;
+      if (TYPE_INT != t)
+      {
+        clear(TYPE_INT);
+      }
+      i = v;
+      return *this;
+    }
 
-      //first clear
-      clear();
-    
-      // then perform actual assignment of members
-      t =  obj.t;
-      
-      switch(t) 
-        {
-        case TYPE_NONE:
-          break;
-        case TYPE_INT:
-          i = obj.i;
-          break;
-        case TYPE_FLOAT:
-          f = obj.f;
-        break;
-        case TYPE_STRING:
-          s = new StringType(*obj.s);
-          break;
-        case TYPE_MAP:
-          m = new MapType(*obj.m);
-          break;
-        case TYPE_LIST:
-          l = new ListType(*obj.l);
-          break;
-        }
+    Element& operator=(bool v) 
+    {
+      if (TYPE_INT != t)
+      {
+        clear(TYPE_INT);
+      }
+      i = v;
+      return *this;
+    }
 
+    Element& operator=(IntType v) 
+    {
+      if (TYPE_INT != t)
+      {
+        clear(TYPE_INT);
+      }
+      i = v;
+      return *this;
+    }
+
+    Element& operator=(float v) 
+    {
+      if (TYPE_FLOAT != t)
+      {
+        clear(TYPE_FLOAT);
+      }
+      f = v;
+      return *this;
+    }
+
+    Element& operator=(FloatType v) 
+    {
+      if (TYPE_FLOAT != t)
+      {
+        clear(TYPE_FLOAT);
+      }
+      f = v;
+      return *this;
+    }
+
+    Element& operator=(PtrType v) 
+    {
+      if (TYPE_PTR != t)
+      {
+        clear(TYPE_PTR);
+      }
+      p = v;
+      return *this;
+    }
+
+    Element& operator=(const char * v) 
+    {
+      if (TYPE_STRING != t || !s->unique())
+      {
+        clear(TYPE_STRING);
+        s = new DataType<StringType>(v);
+      } else {
+        *s = StringType(v);
+      }
+      return *this;
+    }
+
+    Element& operator=(const StringType & v) 
+    {
+      if (TYPE_STRING != t || !s->unique())
+      {
+        clear(TYPE_STRING);
+        s = new DataType<StringType>(v);
+      } else {
+        *s = v;
+      }
+      return *this;
+    }
+
+    Element& operator=(const MapType & v) 
+    {
+      if (TYPE_MAP != t || !m->unique())
+      {
+        clear(TYPE_MAP);
+        m = new DataType<MapType>(v);
+      } else {
+        *m = v;
+      }
+      return *this;
+    }
+
+    Element& operator=(const ListType & v) 
+    {
+      if (TYPE_LIST != t || !l->unique())
+      {
+        clear(TYPE_LIST);
+        l = new DataType<ListType>(v);
+      } else {
+        *l = v;
+      }
       return *this;
     }
 
     /// Check for equality with another Element.
-    bool operator==(const Element& o) const
-    {
-        if (t != o.t) return false;
-        switch(t) {
-            case TYPE_NONE: return true;
-            case TYPE_INT: return i == o.i;
-            case TYPE_FLOAT: return f == o.f;
-            case TYPE_STRING: return *s == *o.s;
-            case TYPE_MAP: return *m == *o.m;
-            case TYPE_LIST: return *l == *o.l;
-        }
-        return false;
-    }
+    bool operator==(const Element& o) const;
 
-    /// Check for inequality with another Element.
-    bool operator!=(const Element m) const
+#if defined(__GNUC__) && __GNUC__ < 3
+    bool operator!=(const Element& o) const
     {
-        return !(*this == m);
+        return !(*this == o);
+    }
+#endif // defined(__GNUC__) && __GNUC__ < 3
+    
+    /// Check for inequality with anything we can check equality with
+    template<class C>
+    bool operator!=(C c) const
+    {
+        return !(*this == c);
     }
 
     /// Check for equality with a int.
@@ -237,17 +295,25 @@ public:
       return (t == TYPE_INT && i == v);
     }
 
-    /// Check for inequality with a int.
-    bool operator!=(IntType v) const { return !(*this == v); }
-
     /// Check for equality with a double.
     bool operator==(FloatType v) const
     {
       return t == TYPE_FLOAT && f == v;
     }
 
-    /// Check for inequality with a double.
-    bool operator!=(FloatType v) const { return !(*this == v); }
+    /// Check for equality with a pointer.
+    bool operator==(PtrType v) const
+    {
+      return t == TYPE_PTR && p == v;
+    }
+
+    /// Check for equality with a const char *.
+    bool operator==(const char * v) const
+    {
+      if(t == TYPE_STRING)
+        return (*s == v);
+      return false;
+    }
 
     /// Check for equality with a std::string.
     bool operator==(const StringType& v) const
@@ -257,9 +323,6 @@ public:
       return false;
     }
 
-    /// Check for inequality with a std::string.
-    bool operator!=(const StringType& v) const { return !(*this == v); }
-
     /// Check for equality with a MapType.
     bool operator==(const MapType& v) const
     {
@@ -268,9 +331,6 @@ public:
       return false;
     }
 
-    /// Check for inequality with a MapType.
-    bool operator!=(const MapType& v) const { return !(*this == v); }
-
     /// Check for equality with a ListType.
     bool operator==(const ListType& v) const
     {
@@ -279,9 +339,6 @@ public:
       return false;
     }
 
-    /// Check for inequality with a ListType.
-    bool operator!=(const ListType& v) const { return !(*this == v); }
-
     /// Get the current type.
     Type getType() const { return t; }
     /// Check whether the current type is nothing.
@@ -290,6 +347,8 @@ public:
     bool isInt() const { return (t == TYPE_INT); }
     /// Check whether the current type is double.
     bool isFloat() const { return (t == TYPE_FLOAT); }
+    /// Check whether the current type is pointer.
+    bool isPtr() const { return (t == TYPE_PTR); }
     /// Check whether the current type is numeric.
     bool isNum() const { return ((t == TYPE_FLOAT) || (t == TYPE_INT)); }
     /// Check whether the current type is std::string.
@@ -300,17 +359,35 @@ public:
     bool isList() const { return (t == TYPE_LIST); }
 
     /// Retrieve the current value as a int.
-    long asInt() const throw (WrongTypeException)
+    IntType asInt() const throw (WrongTypeException)
     {
         if (t == TYPE_INT) return i;
         throw WrongTypeException();
     }
+    IntType Int() const
+    {
+        return i;
+    }
     /// Retrieve the current value as a double.
     FloatType asFloat() const throw (WrongTypeException)
     {
         if (t == TYPE_FLOAT) return f;
         throw WrongTypeException();
     }
+    FloatType Float() const
+    {
+        return f;
+    }
+    /// Retrieve the current value as a pointer.
+    PtrType asPtr() const throw (WrongTypeException)
+    {
+        if (t == TYPE_PTR) return p;
+        throw WrongTypeException();
+    }
+    PtrType Ptr() const
+    {
+        return p;
+    }
     /// Retrieve the current value as a number.
     FloatType asNum() const throw (WrongTypeException)
     {
@@ -327,9 +404,17 @@ public:
     /// Retrieve the current value as a non-const std::string reference.
     std::string& asString() throw (WrongTypeException)
     {
-        if (t == TYPE_STRING) return *s;
+        if (t == TYPE_STRING) return *(s = s->makeUnique());
         throw WrongTypeException();
     }
+    const StringType& String() const
+    {
+        return *s;
+    }
+    StringType& String()
+    {
+        return *(s = s->makeUnique());
+    }
     /// Retrieve the current value as a const MapType reference.
     const MapType& asMap() const throw (WrongTypeException)
     {
@@ -339,9 +424,17 @@ public:
     /// Retrieve the current value as a non-const MapType reference.
     MapType& asMap() throw (WrongTypeException)
     {
-        if (t == TYPE_MAP) return *m;
+        if (t == TYPE_MAP) return *(m = m->makeUnique());
         throw WrongTypeException();
     }
+    const MapType& Map() const
+    {
+        return *m;
+    }
+    MapType& Map()
+    {
+        return *(m = m->makeUnique());
+    }
     /// Retrieve the current value as a const ListType reference.
     const ListType& asList() const throw (WrongTypeException)
     {
@@ -351,29 +444,67 @@ public:
     /// Retrieve the current value as a non-const ListType reference.
     ListType& asList() throw (WrongTypeException)
     {
-        if (t == TYPE_LIST) return *l;
+        if (t == TYPE_LIST) return *(l = l->makeUnique());
         throw WrongTypeException();
     }
+    const ListType& List() const
+    {
+        return *l;
+    }
+    ListType& List()
+    {
+        return *(l = l->makeUnique());
+    }
+
+    static const char * typeName(Type);
 
 protected:
 
+    template<class C>
+    class DataType
+    {
+    public:
+        DataType() : _refcount(1) {}
+        DataType(const C& c) : _refcount(1), _data(c) {}
+
+        DataType& operator=(const C& c) {_data = c; return *this;}
+
+        bool operator==(const C& c) const {return _data == c;}
+
+        void ref() {++_refcount;}
+        void unref() {if(--_refcount == 0) delete this;}
+
+        bool unique() const {return _refcount == 1;}
+        DataType* makeUnique()
+        {
+           if(unique())
+               return this;
+           unref(); // _refcount > 1, so this is fine
+           return new DataType(_data);
+        }
+
+        operator C&() {return _data;}
+//        operator const C&() const {return _data;}
+
+    private:
+        DataType(const DataType&); // unimplemented
+        DataType& operator=(const DataType&); // unimplemented
+
+        unsigned long _refcount;
+        C _data;
+    };
+
     Type t;
     union {
       IntType i;
       FloatType f;
-      StringType* s;
-      MapType* m;
-      ListType* l;
-      void* n;
+      void* p;
+      DataType<StringType>* s;
+      DataType<MapType>* m;
+      DataType<ListType>* l;
     };
 };
 
-typedef Element::IntType IntType;
-typedef Element::FloatType FloatType;
-typedef Element::StringType StringType;
-typedef Element::MapType MapType;
-typedef Element::ListType ListType;
-
 } } // namespace Atlas::Message
 
 
