@@ -23,10 +23,10 @@ template <class Decoder>
 class Layer : public Encoder, public Decoder
 {
 public:
-    Layer(Atlas::Bridge* b) : Encoder(b), Decoder() {}
-    virtual ~Layer() {}
+    Layer(Atlas::Bridge* b) : Encoder(b), Decoder() { }
+    virtual ~Layer() { }
 };
 
-} } // Atlas::Message namespace
+} } // namespace Atlas::Message
 
 #endif
