@@ -15,28 +15,29 @@ namespace Atlas { namespace Objects {
  * This Encoder can be used to send objects in the Atlas::Objects hierarchy to
  * a certain Bridge (e.g. a codec).
  *
- * Simply call the StreamMessage member with a pointer to the object to be
+ * Simply call the streamMessage member with a pointer to the object to be
  * sent.
  *
  * @see Atlas::Objects::Decoder
  * @author Stefanus Du Toit <sdt@gmx.net>
  */
-class Encoder : public Atlas::EncoderBase
+class ObjectsEncoder : public Atlas::EncoderBase
 {
 public:
     /// The default constructor.
     /// @param b The Bridge to which objects are to be sent.
-    Encoder(Atlas::Bridge* b) : EncoderBase(b) { }
+    explicit ObjectsEncoder(Atlas::Bridge & b) : EncoderBase(b) { }
     /// The default destructor.
-    virtual ~Encoder() { }
+    ~ObjectsEncoder();
 
     /// Send an object to the bridge.
     /// @param o The object that is to be sent.
-    virtual void streamMessage(const Atlas::Objects::Root* o)
+    template <class ObjectData>
+    void streamObjectsMessage(const Atlas::Objects::SmartPtr<ObjectData> & o)
     {
-        b->streamMessage(Bridge::MapBegin);
-        o->sendContents(b);
-        b->mapEnd();
+        m_b.streamMessage();
+        o->sendContents(m_b);
+        m_b.mapEnd();
     }
 };
 
